import os
from .annotation import log_param, log_metric, trace
from .onnx_client import OnnxClient

os.environ["GIT_PYTHON_REFRESH"] = "quiet"

__all__ = [
    "log_param",
    "log_metric",
    "trace",
    "OnnxClient",
]



import logging
import os
import pika
import onnx

from abc import ABCMeta, abstractmethod
from mlflow.models.signature import ModelSignature
from mlflow.models.utils import ModelInputExample
from .env_checker import check_env_vars

"""
ModelInputExample
    mlflow 1.x 버전에서는 from mlflow.models.utils import ModelInputExample
    mlflow 2.x 버전에서는 from mlflow.models.signature import ModelInputExample
"""

logger = logging.getLogger(__name__)


class BaseMLflowClient(metaclass=ABCMeta):
    isProduction: bool
    ONNX_TO_TRITON_DTYPE = {
        onnx.TensorProto.BOOL: "TYPE_BOOL",
        onnx.TensorProto.UINT8: "TYPE_UINT8",
        onnx.TensorProto.UINT16: "TYPE_UINT16",
        onnx.TensorProto.UINT32: "TYPE_UINT32",
        onnx.TensorProto.UINT64: "TYPE_UINT64",
        onnx.TensorProto.INT8: "TYPE_INT8",
        onnx.TensorProto.INT16: "TYPE_INT16",
        onnx.TensorProto.INT32: "TYPE_INT32",
        onnx.TensorProto.INT64: "TYPE_INT64",
        onnx.TensorProto.FLOAT16: "TYPE_FP16",
        onnx.TensorProto.FLOAT: "TYPE_FP32",
        onnx.TensorProto.DOUBLE: "TYPE_FP64",
        onnx.TensorProto.STRING: "TYPE_STRING",
        # Brain floating point (bfloat16) is not directly supported in ONNX, you might need to handle it separately
        # "TYPE_BF16": <corresponding ONNX type if available or a custom handler>
    }

    def __init__(self):
        profile = os.environ.get("PROFILE")
        if profile not in ["prod", "production"]:
            logging.warn(
                """
                You are using a non-production profile.
                If you are in production, please set the PROFILE environment variable to 'prod' or 'production'.
                """
            )
            self.isProduction = False

            if os.environ.get("MLFLOW_S3_ENDPOINT_URL") is None:
                os.environ["MLFLOW_S3_ENDPOINT_URL"] = "http://localhost:9000"
            if os.environ.get("MLFLOW_TRACKING_URI") is None:
                os.environ["MLFLOW_TRACKING_URI"] = "http://localhost:5000"
            if os.environ.get("AWS_ACCESS_KEY_ID") is None:
                os.environ["AWS_ACCESS_KEY_ID"] = "minio"
            if os.environ.get("AWS_SECRET_ACCESS_KEY") is None:
                os.environ["AWS_SECRET_ACCESS_KEY"] = "miniostorage"

            self.model_name = os.environ.get("MODEL_NAME", "my_model")
            self.train_id = os.environ.get("TRAIN_ID", "1")
            return
        else:
            self.isProduction = True
            check_env_vars()

            self.model_name = os.environ["MODEL_NAME"]
            self.train_id = os.environ["TRAIN_ID"]
            self._uploadModelExchange = os.environ["RABBIT_MODEL_UPLOAD_TOPIC"]
            # self._connection = pika.BlockingConnection(
            #     pika.URLParameters(os.environ["RABBIT_ENDPOINT_URL"])
            # )

    def get_connection(self):
        return pika.BlockingConnection(
            pika.URLParameters(os.environ["RABBIT_ENDPOINT_URL"])
        )

    def get_triton_compatible_type(self, tensor_type):
        return self.ONNX_TO_TRITON_DTYPE.get(tensor_type.elem_type, "UNKNOWN")

    @abstractmethod
    def upload(
        self,
        model,
        # input_example: Schema,
        # output_example: Schema,
    ):
        pass

    @abstractmethod
    def _log_tensor(self, model):
        """
        _log_tensor:
        onnx.load 로 로드된 모델을 통해 input tensor와 output tensor의 정보를 출력합니다.
        """
        pass

    @abstractmethod
    def _log_model(
        self,
        model,
        signature: ModelSignature,
        # input_example: ModelInputExample,
    ) -> str:
        pass



import os
import logging

logger = logging.getLogger(__name__)


def check_env_vars(
    required: list = [
        "MLFLOW_S3_ENDPOINT_URL",
        "MLFLOW_TRACKING_URI",
        "AWS_ACCESS_KEY_ID",
        "AWS_SECRET_ACCESS_KEY",
        "RABBIT_ENDPOINT_URL",
        "RABBIT_MODEL_UPLOAD_TOPIC",
        "TRAIN_ID",
        "MODEL_NAME",
    ]
):
    missing_vars = [var for var in required if not os.environ.get(var)]

    logger.info("Required environment variables: %s", required)
    if missing_vars:
        error_message = (
            f"다음 환경 변수들이 설정되지 않았습니다: {', '.join(missing_vars)}"
        )
        raise EnvironmentError(error_message)



import functools
import mlflow
import logging
import threading
import os
import onnx

from typing import Optional
from pathlib import Path
from .onnx_client import OnnxClient

# 스레드 로컬 스토리지 설정
_thread_local = threading.local()
client = OnnxClient()
logger = logging.getLogger(__name__)


def set_log_functions(log_param, log_metric):
    _thread_local.log_param = log_param
    _thread_local.log_metric = log_metric


def clear_log_functions():
    _thread_local.log_param = None
    _thread_local.log_metric = None


def log_param(key, value):
    if hasattr(_thread_local, "log_param") and _thread_local.log_param:
        _thread_local.log_param(key, value)
    else:
        raise RuntimeError(
            "trace 데코레이터 내에서 log_param 함수가 사용되지 않았습니다."
        )


def log_metric(key, value, step=None):
    if hasattr(_thread_local, "log_metric") and _thread_local.log_metric:
        _thread_local.log_metric(key, value, step)
    else:
        raise RuntimeError(
            "trace 데코레이터 내에서 log_metric 함수가 사용되지 않았습니다."
        )


def trace(experiment_name: Optional[str] = None, run_name: Optional[str] = None):
    """
    MLflow 실험을 설정하고 함수 실행 결과를 ONNX 모델로 로깅하는 데코레이터.

    Args:
        experiment_name (str): MLflow 실험 이름.
        run_name (Optional[str]): MLflow 실행 이름. Defaults to None.
    """

    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            # 실험 설정
            if experiment_name is not None:
                mlflow.set_experiment(experiment_name)
            with mlflow.start_run(run_name=run_name):
                try:
                    # 로깅 함수 설정
                    set_log_functions(
                        log_param=mlflow.log_param,
                        log_metric=mlflow.log_metric,
                    )
                    logger.info(
                        f"Logging functions have been set for experiment '{experiment_name}'."
                    )

                    # 함수 실행
                    result = func(*args, **kwargs)

                    # 반환값이 str 또는 Path인지 확인
                    if isinstance(result, (str, os.PathLike)):
                        model_path = Path(result)  # pathlib.Path로 변환
                        if not model_path.exists():
                            raise FileNotFoundError(
                                f"Model path does not exist: {model_path}"
                            )

                        logger.info(f"Loading ONNX model from: {model_path}")

                        # # ONNX 모델 로드
                        # onnx_model = onnx.load(model_path)

                        # 모델 로깅
                        client.upload(model_path)

                        logger.info(f"Model logged successfully from: {model_path}")
                    else:
                        raise TypeError(
                            f"Function '{func.__name__}' must return a str or Path, but got {type(result)}"
                        )
                    return result  # 원래 함수의 반환값을 그대로 반환
                except Exception as e:
                    logger.error(f"Error in trace decorator: {e}")
                    raise e
                finally:
                    # 로깅 함수 해제
                    clear_log_functions()
                    logger.info("Logging functions have been cleared.")

        return wrapper

    return decorator
